from enum import Enum
from typing import Any, Optional

from pydantic import BaseModel, model_serializer, RootModel


class EnumModel(Enum):
    def __new__(cls, *args):
        obj = object.__new__(cls)
        obj._value_ = args[0]
        return obj

    def __str__(self) -> str:
        return str(self._value_)

    def __repr__(self) -> str:
        return str(self._value_)

    def __eq__(self, obj) -> bool:
        return isinstance(obj, self.__class__) and self.value == obj.value

    def __hash__(self):
        return hash(self.value)

    @classmethod
    def from_string(cls, value: str):
        for item in cls.__members__.values():
            if item.value == value:
                return item
        raise ValueError(f"No enum value {value} found.")

    @model_serializer
    def serialize(self):
        return self.value


class LHBaseModel(BaseModel):
    def write_to_file(self, filename: str, indent: Optional[int] = None) -> None:
        with open(filename, "w", encoding="utf-8") as file_pointer:
            file_pointer.write(self.model_dump_json(indent=indent))

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))


class ListModel(RootModel):
    root: list[Any]

    def append(self, item):
        self.root.append(item)

    def extend(self, items):
        self.root.extend(items)

    def insert(self, index, item):
        self.root.insert(index, item)

    def remove(self, item):
        self.root.remove(item)

    def pop(self, index=-1):
        return self.root.pop(index)

    def clear(self):
        self.root.clear()

    def index(self, item, start=0, end=None):
        return self.root.index(item, start, end)

    def count(self, item):
        return self.root.count(item)

    def sort(self, key=None, reverse=False):
        self.root.sort(key=key, reverse=reverse)

    def reverse(self):
        self.root.reverse()

    def __getitem__(self, index):
        return self.root[index]

    def __setitem__(self, index, value):
        self.root[index] = value

    def __delitem__(self, index):
        del self.root[index]

    def __len__(self):
        return len(self.root)

    def __iter__(self):
        return iter(self.root)

    def __contains__(self, item):
        return item in self.root

    def write_to_file(self, filename: str, indent: Optional[int] = None) -> None:
        with open(filename, "w", encoding="utf-8") as file_pointer:
            file_pointer.write(self.model_dump_json(indent=indent))


class DictModel(RootModel):
    root: dict[str, Any]

    def items(self):
        return self.root.items()

    def keys(self):
        return self.root.keys()

    def values(self):
        return self.root.values()

    def __getitem__(self, key: str) -> Any:
        return self.root[key]

    def __setitem__(self, key: str, newvalue: Any):
        self.root[key] = newvalue

    def __contains__(self, key: str):
        return key in self.root

    def __delitem__(self, key: str):
        del self.root[key]

    def dict(self, *args, **kwargs):
        ret = super().dict(*args, **kwargs)
        if "root" in ret:
            ret = ret["root"]
        return ret

    def write_to_file(self, filename: str, indent: Optional[int] = None) -> None:
        with open(filename, "w", encoding="utf-8") as file_pointer:
            file_pointer.write(self.model_dump_json(indent=indent))



from __future__ import annotations

from os import listdir, path, stat
from shutil import rmtree
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

import pandas as pd
from sklearn.feature_selection import SelectFromModel
from sklearn.metrics import accuracy_score

from learninghouse.api.errors.brain import (
    BrainBadRequest,
    BrainExists,
    BrainNoConfiguration,
    BrainNotActual,
    BrainNotEnoughData,
    BrainNotTrained,
)
from learninghouse.core.logging import logger
from learninghouse.core.settings import service_settings
from learninghouse.models.brain import (
    Brain,
    BrainConfiguration,
    BrainDeleteResult,
    BrainEstimatorType,
    BrainFileType,
    BrainInfo,
    BrainInfos,
    BrainPredictionResult,
)
from learninghouse.services.preprocessing import DatasetPreprocessing

if TYPE_CHECKING:
    from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor


class BrainService:
    brains: Dict[str, Dict[str, Union[int, Brain]]] = {}

    @classmethod
    def list_all(cls) -> BrainInfos:
        brains: Dict[str, BrainInfo] = {}
        for directory in listdir(service_settings().brains_directory):
            try:
                brains[directory] = cls.get_info(directory)
            except BrainNoConfiguration:
                pass

        return BrainInfos.parse_obj(brains)

    @staticmethod
    def get_info(name: str) -> BrainInfo:
        info: Optional[BrainInfo] = None
        if Brain.is_trained(name):
            try:
                info = Brain.load_trained(name).info
            except (AttributeError, BrainNotTrained):
                pass

        if info is None:
            if BrainConfiguration.json_config_file_exists(name):
                info = Brain(name).info
                training_data_file = Brain.sanitize_filename(
                    name, BrainFileType.TRAINING_DATA_FILE
                )
                if path.exists(training_data_file):
                    data = pd.read_csv(training_data_file)
                    info.training_data_size = len(data.index)
            else:
                raise BrainNoConfiguration(name)

        return info

    @classmethod
    def request(
        cls,
        name: str,
        dependent_value: Optional[Any] = None,
        sensors_data: Optional[Dict[str, Any]] = None
    ) -> BrainInfo:
        filename = Brain.sanitize_filename(
            name, BrainFileType.TRAINING_DATA_FILE)

        trainings_data: Optional[Dict[str, Any]] = sensors_data

        if sensors_data is not None:
            if dependent_value is not None:
                trainings_data[name] = dependent_value
            else:
                raise BrainBadRequest("Missing dependent variable!")

        if trainings_data is None:
            if path.exists(filename):
                data = pd.read_csv(filename)
            else:
                raise BrainNotEnoughData()
        else:
            logger.debug(trainings_data)
            trainings_data = DatasetPreprocessing.add_time_information(
                trainings_data)
            if path.exists(filename):
                data_temp = pd.read_csv(filename)
                df_new_row = pd.DataFrame([trainings_data])
                data = pd.concat([data_temp, df_new_row], ignore_index=True)
            else:
                data = pd.DataFrame([trainings_data])

            data.to_csv(filename, sep=",", index=False)

        return cls.train(name, data)

    @staticmethod
    def train(name: str, data: pd.DataFrame) -> BrainInfo:
        try:
            brain = Brain(name)

            if len(data.index) < 10:
                raise BrainNotEnoughData()

            (
                brain,
                x_train,
                x_test,
                y_train,
                y_test,
            ) = DatasetPreprocessing.prepare_training(brain, data, False)

            estimator = brain.estimator()

            selector = SelectFromModel(estimator)
            selector.fit(x_train, y_train)

            brain.dataset.features = x_train.columns[
                (selector.get_support())
            ].values.tolist()

            (
                brain,
                x_train,
                x_test,
                y_train,
                y_test,
            ) = DatasetPreprocessing.prepare_training(brain, data, True)

            estimator.fit(x_train, y_train)

            if BrainEstimatorType.CLASSIFIER == brain.configuration.estimator.typed:
                y_pred = estimator.predict(x_test)
                score = accuracy_score(y_test, y_pred)
            else:
                score = estimator.score(x_test, y_test)

            brain.store_trained(x_train.columns.tolist(),
                                len(data.index), score)

            return brain.info
        except FileNotFoundError as exc:
            raise BrainNoConfiguration(name) from exc

    @classmethod
    def prediction(cls, name: str, request_data: Dict[str, Any]):
        try:
            brain = cls.load_brain(name)
            if not brain.actual_versions:
                raise BrainNotActual(name, brain.version)

            request_data = DatasetPreprocessing.add_time_information(
                request_data)

            data = pd.DataFrame([request_data])
            prepared_data = DatasetPreprocessing.prepare_prediction(
                brain, data)

            prediction = brain.estimator().predict(prepared_data)

            if (
                brain.configuration.dependent_encode
                and brain.configuration.estimator.typed == BrainEstimatorType.CLASSIFIER
            ):
                prediction = brain.dataset.dependent_encoder.inverse_transform(
                    prediction
                )
                prediction = list(map(bool, prediction))
            else:
                prediction = list(map(float, prediction))

            return BrainPredictionResult(
                brain=brain.info,
                preprocessed=prepared_data.head(1).to_dict("records")[0],
                prediction=prediction[0],
            )
        except FileNotFoundError as exc:
            raise BrainNotTrained(name) from exc

    @classmethod
    def load_brain(cls, name: str) -> Brain:
        filename = Brain.sanitize_filename(name, BrainFileType.TRAINED_FILE)
        stamp = stat(filename).st_mtime

        if not (name in cls.brains and cls.brains[name]["stamp"] == stamp):
            cls.brains[name] = {"stamp": stamp,
                                "brain": Brain.load_trained(name)}

        return cls.brains[name]["brain"]


class BrainConfigurationService:
    @staticmethod
    def get(name: str) -> BrainConfiguration:
        try:
            return BrainConfiguration.from_json_file(name)
        except FileNotFoundError as exc:
            raise BrainNoConfiguration(name) from exc

    @staticmethod
    def create(configuration: BrainConfiguration) -> BrainConfiguration:
        if BrainConfiguration.json_config_file_exists(configuration.name):
            raise BrainExists(configuration.name)

        configuration.to_json_file(configuration.name)

        return configuration

    @staticmethod
    def update(name: str, configuration: BrainConfiguration) -> BrainConfiguration:
        if not BrainConfiguration.json_config_file_exists(name):
            raise BrainNoConfiguration(name)

        configuration.to_json_file(name)

        return configuration

    @staticmethod
    def delete(name: str) -> BrainDeleteResult:
        brainpath = Brain.sanitize_directory(name)

        if not path.exists(brainpath):
            raise BrainNoConfiguration(name)

        logger.info(f"Remove brain: {name}")
        rmtree(brainpath)

        return BrainDeleteResult(name=name)



from typing import Optional

from fastapi import __version__ as fastapi_version
from jose import __version__ as jose_version
from loguru import __version__ as loguru_version
from numpy.version import version as np_version
from pandas import __version__ as pd_version
from passlib import __version__ as passlib_version
from pydantic.version import VERSION as pydantic_version
from sklearn import __version__ as skl_version
from uvicorn import __version__ as uvicorn_version

from learninghouse.models import LearningHouseVersions

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

versions = LearningHouseVersions(
    service=__version__,
    fastapi=fastapi_version,
    uvicorn=uvicorn_version,
    pydantic=pydantic_version,
    sklearn=skl_version,
    numpy=np_version,
    pandas=pd_version,
    jose=jose_version,
    passlib=passlib_version,
    loguru=loguru_version,
)



from typing import Any, Dict, List, Optional

from fastapi import Request, status
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse

from learninghouse.core.logging import logger
from learninghouse.models import (
    LearningHouseErrorMessage,
    LearningHouseValidationErrorMessage,
)

MIMETYPE_JSON = "application/json"


class LearningHouseException(Exception):
    STATUS_CODE = status.HTTP_500_INTERNAL_SERVER_ERROR
    UNKNOWN = "UNKNOWN"
    DESCRIPTION = "An unknown exception occurred " + "while handling your request."

    def __init__(
        self,
        status_code: Optional[int] = None,
        key: Optional[str] = None,
        description: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None,
    ):
        super().__init__()
        self.http_status_code: int = status_code or self.STATUS_CODE
        self.error: LearningHouseErrorMessage = LearningHouseErrorMessage(
            error=key or self.UNKNOWN, description=description or self.DESCRIPTION
        )
        self.headers = headers if headers else {}

    def response(self) -> JSONResponse:
        return JSONResponse(
            content=self.error.dict(),
            status_code=self.http_status_code,
            headers=self.headers,
        )

    @classmethod
    def api_description(cls) -> Dict:
        return {
            "model": LearningHouseErrorMessage,
            "description": "An exception occured which is not handled by the service now. "
            + "Please write an issue on GitHub.",
            "content": {
                MIMETYPE_JSON: {
                    "example": {"error": cls.UNKNOWN, "description": cls.DESCRIPTION}
                }
            },
        }


class LearningHouseSecurityException(LearningHouseException):
    STATUS_CODE = status.HTTP_403_FORBIDDEN
    SECURITY_EXCEPTION = "SECURITY_EXCEPTION"
    DESCRIPTION = "A security violation occured while handling your request."

    def __init__(self, description: str, key: Optional[str] = None):
        key = key or self.SECURITY_EXCEPTION

        super().__init__(self.STATUS_CODE, key, description or self.DESCRIPTION)

    @classmethod
    def api_description(cls) -> Dict:
        return {
            "model": LearningHouseErrorMessage,
            "description": "The request didn't pass security checks.",
            "content": {
                MIMETYPE_JSON: {
                    "example": {
                        "error": cls.SECURITY_EXCEPTION,
                        "description": cls.DESCRIPTION,
                    }
                }
            },
        }


class LearningHouseUnauthorizedException(LearningHouseException):
    STATUS_CODE = status.HTTP_401_UNAUTHORIZED
    UNAUTHORIZED = "UNAUTHORIZED"
    DESCRIPTION = "Could not validate credentials"

    def __init__(self, description: Optional[str] = None):
        super().__init__(
            self.STATUS_CODE,
            self.UNAUTHORIZED,
            description or self.DESCRIPTION,
            {"WWW-Authenticate": "Bearer"},
        )

    @classmethod
    def api_description(cls) -> Dict:
        return {
            "model": LearningHouseErrorMessage,
            "description": "The request didn't pass security checks.",
            "content": {
                MIMETYPE_JSON: {
                    "example": {
                        "error": cls.UNAUTHORIZED,
                        "description": cls.DESCRIPTION,
                    }
                }
            },
        }


class LearningHouseValidationError(LearningHouseException):
    STATUS_CODE = status.HTTP_422_UNPROCESSABLE_ENTITY
    VALIDATION_ERROR = "VALIDATION_ERROR"
    DESCRIPTION = "A validation error occurred while handling your request."

    def __init__(self, error: Optional[RequestValidationError] = None):
        super().__init__(
            self.STATUS_CODE, self.VALIDATION_ERROR, str(error) or self.DESCRIPTION
        )
        self.validations: List[Dict[str, Any]] = error.errors()

    @classmethod
    def api_description(cls) -> Dict:
        return {
            "model": LearningHouseValidationErrorMessage,
            "description": "The request didn't pass input validation",
            "content": {
                MIMETYPE_JSON: {
                    "example": {
                        "error": cls.VALIDATION_ERROR,
                        "description": cls.DESCRIPTION,
                        "validations": [
                            {
                                "loc": ["body", "..."],
                                "msg": "example",
                                "type": "value...",
                                "ctx": {},
                            }
                        ],
                    }
                }
            },
        }

    def response(self) -> JSONResponse:
        validation_error = LearningHouseValidationErrorMessage.from_error_message(
            self.error, self.validations
        )
        return JSONResponse(
            content=validation_error.dict(), status_code=self.http_status_code
        )


async def validation_error_handler(
    _: Request, exc: RequestValidationError
) -> JSONResponse:
    return LearningHouseValidationError(exc).response()


async def learninghouse_exception_handler(_: Request, exc: LearningHouseException):
    response = exc.response()

    if isinstance(
        exc, (LearningHouseSecurityException, LearningHouseUnauthorizedException)
    ):
        logger.warning(exc.error.description)

    return response



from typing import Dict

from fastapi import status

from learninghouse.api.errors import LearningHouseException
from learninghouse.models import LearningHouseErrorMessage

MIMETYPE_JSON = "application/json"


class NoSensor(LearningHouseException):
    STATUS_CODE = status.HTTP_404_NOT_FOUND
    NO_SENSOR = "NO_SENSOR"
    DESCRIPTION = "No sensor {name} found."

    def __init__(self, name: str):
        super().__init__(
            self.STATUS_CODE, self.NO_SENSOR, self.DESCRIPTION.format(name=name)
        )

    @classmethod
    def api_description(cls) -> Dict:
        return {
            "model": LearningHouseErrorMessage,
            "description": "No sensor with given name found",
            "content": {
                MIMETYPE_JSON: {
                    "example": {"error": cls.NO_SENSOR, "description": cls.DESCRIPTION}
                }
            },
        }


class SensorExists(LearningHouseException):
    STATUS_CODE = status.HTTP_400_BAD_REQUEST
    EXISTS = "EXISTS"
    DESCRIPTION = "The existing sensor {name} can not be recreated. Use PUT to update."

    def __init__(self, name: str):
        super().__init__(
            self.STATUS_CODE, self.EXISTS, self.DESCRIPTION.format(name=name)
        )

    @classmethod
    def api_description(cls) -> Dict:
        return {
            "model": LearningHouseErrorMessage,
            "description": "An existing sensor can not be recreated. Use PUT to update.",
            "content": {
                MIMETYPE_JSON: {
                    "example": {"error": cls.EXISTS, "description": cls.DESCRIPTION}
                }
            },
        }



import logging
import sys
from pprint import pformat
from types import FrameType
from typing import cast

from loguru import logger
from loguru._defaults import LOGURU_FORMAT

from learninghouse.models.base import EnumModel


class LoggingLevelEnum(EnumModel):
    DEBUG = "DEBUG", logging.DEBUG
    INFO = "INFO", logging.INFO
    WARNING = "WARNING", logging.WARNING
    ERROR = "ERROR", logging.ERROR
    CRITICAL = "CRITICAL", logging.CRITICAL

    def __init__(self, description: str, level: int):
        # pylint: disable=super-init-not-called
        self._description: str = description
        self._level: int = level

    @property
    def description(self) -> str:
        return self._description

    @property
    def level(self) -> int:
        return self._level


class LoggingHandler(logging.Handler):
    def emit(self, record: logging.LogRecord) -> None:  # pragma: no cover
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = str(record.levelno)

        frame, depth = logging.currentframe(), 2
        while frame.f_code.co_filename == logging.__file__:  # noqa: WPS609
            frame = cast(FrameType, frame.f_back)
            depth += 1

        log = logger.bind(request_id="app")
        log.opt(depth=depth, exception=record.exc_info).log(
            level,
            record.getMessage(),
        )


def format_record(record: dict) -> str:
    format_string = LOGURU_FORMAT
    if record["extra"].get("payload") is not None:
        record["extra"]["payload"] = pformat(
            record["extra"]["payload"], indent=4, compact=True, width=88
        )
        format_string += "\n<level>{extra[payload]}</level>"

    format_string += "{exception}\n"
    return format_string


def initialize_logging(logging_level: LoggingLevelEnum) -> None:
    logging_handler = LoggingHandler(level=logging_level.level)

    logging.basicConfig(handlers=[logging_handler], level=0)

    uvicorn_logger = logging.getLogger("uvicorn")
    uvicorn_logger.propagate = False

    for uvicorn_logger_name in ("fastapi", "uvicorn.asgi", "uvicorn.access"):
        logging.getLogger(uvicorn_logger_name).handlers = [logging_handler]

    logger.bind(request_id=None, method=None)

    logger.configure(
        handlers=[
            {"sink": sys.stdout, "level": logging_level.level, "format": format_record}
        ]
    )



from __future__ import annotations

import json
from os import path
from typing import List

from pydantic import Field

from learninghouse.core.logging import logger
from learninghouse.core.settings import service_settings
from learninghouse.models.base import (
    EnumModel,
    LHBaseModel,
    ListModel,
)


class SensorType(EnumModel):
    NUMERICAL = "numerical"
    CATEGORICAL = "categorical"

    def __init__(self, typed: str):
        # pylint: disable=super-init-not-called
        self._typed: str = typed

    @property
    def typed(self) -> str:
        return self._typed


class SensorConfiguration(LHBaseModel):
    name: str = Field(None, example="azimuth")
    typed: SensorType = Field(None, example=SensorType.NUMERICAL)


class Sensor(SensorConfiguration):
    pass


class Sensors(ListModel):
    root: List[Sensor] = Field(
        None,
        example=[
            {"name": "azimuth", "typed": "numerical"},
            {"name": "elevation", "typed": "numerical"},
            {"name": "rain_gauge", "typed": "numerical"},
            {"name": "pressure", "typed": "numerical"},
            {"name": "pressure_trend_1h", "typed": "categorical"},
            {"name": "temperature_outside", "typed": "numerical"},
            {"name": "temperature_trend_1h", "typed": "categorical"},
            {"name": "light_state", "typed": "categorical"},
        ],
    )

    @classmethod
    def load_config(cls) -> Sensors:
        filename = service_settings().brains_directory / "sensors.json"
        sensors = []

        if path.exists(filename):
            with open(filename, "r", encoding="utf-8") as sensorfile:
                sensors = json.load(sensorfile)
        else:
            logger.warning("No sensors.json found")

        return Sensors.parse_obj(sensors)

    def write_config(self) -> None:
        filename = service_settings().brains_directory / "sensors.json"
        self.write_to_file(filename, indent=4)

    @property
    def numericals(self) -> List[str]:
        return list(
            map(
                lambda x: x.name,
                filter(lambda x: x.typed == SensorType.NUMERICAL, self),
            )
        )

    @property
    def categoricals(self) -> List[str]:
        return list(
            map(
                lambda x: x.name,
                filter(lambda x: x.typed == SensorType.CATEGORICAL, self),
            )
        )


class SensorDeleteResult(LHBaseModel):
    name: str = Field(None, example="azimuth")



from os import path
from pathlib import Path

from fastapi import APIRouter
from fastapi.responses import FileResponse, PlainTextResponse, RedirectResponse

from learninghouse.api.errors import LearningHouseSecurityException
from learninghouse.core.settings import service_settings

settings = service_settings()

router = APIRouter(include_in_schema=False)

UI_DIRECTORY = str(Path(__file__).parent.parent / "ui")


def is_ui_installed() -> bool:
    return path.exists(UI_DIRECTORY)


if is_ui_installed():
    ASSETS_DIRECTORY = f"{UI_DIRECTORY}/assets"

    env_js_content: str = ""

    with open(
        f"{ASSETS_DIRECTORY}/env.template.js", "r", encoding="utf-8"
    ) as template_file:
        env_js_content = template_file.read()
        env_js_content = env_js_content.replace(
            "${LEARNINGHOUSE_API_URL}", f"{settings.base_url_calculated}/api"
        )

    @router.get("/")
    async def redirect_root():
        return RedirectResponse("/ui")

    @router.get("/ui{ui_path:path}")
    async def get_ui(ui_path: str):
        if ui_path and ui_path != "/":
            fullpath = path.normpath(f"{UI_DIRECTORY}{ui_path}")
            if not fullpath.startswith(UI_DIRECTORY):
                raise LearningHouseSecurityException(
                    "Requested file name breaks directory structure"
                )

            if ui_path == "/assets/env.js":
                return PlainTextResponse(env_js_content)

            if not path.exists(fullpath):
                fullpath = f"{UI_DIRECTORY}/index.html"
        else:
            fullpath = f"{UI_DIRECTORY}/index.html"

        return FileResponse(fullpath)



from typing import Dict

from fastapi import status

from learninghouse.api.errors import (
    LearningHouseException,
    LearningHouseSecurityException,
)
from learninghouse.models import LearningHouseErrorMessage

MIMETYPE_JSON = "application/json"


class APIKeyExists(LearningHouseException):
    STATUS_CODE = status.HTTP_400_BAD_REQUEST
    EXISTS = "APIKEY_EXISTS"
    DESCRIPTION = (
        "The existing api key {description} can not be recreated. Use DELETE recreate."
    )

    def __init__(self, description: str):
        super().__init__(
            self.STATUS_CODE,
            self.EXISTS,
            self.DESCRIPTION.format(description=description),
        )

    @classmethod
    def api_description(cls) -> Dict:
        return {
            "model": LearningHouseErrorMessage,
            "description": "An existing api key can not be recreated. Use DELETE and recreate.",
            "content": {
                MIMETYPE_JSON: {
                    "example": {"error": cls.EXISTS, "description": cls.DESCRIPTION}
                }
            },
        }


class NoAPIKey(LearningHouseException):
    STATUS_CODE = status.HTTP_404_NOT_FOUND
    NO_APIKEY = "NO_APIKEY"
    DESCRIPTION = "No API key {description} found."

    def __init__(self, description: str):
        super().__init__(
            self.STATUS_CODE,
            self.NO_APIKEY,
            self.DESCRIPTION.format(description=description),
        )

    @classmethod
    def api_description(cls) -> Dict:
        return {
            "model": LearningHouseErrorMessage,
            "description": "No API key with given description found",
            "content": {
                MIMETYPE_JSON: {
                    "example": {"error": cls.NO_APIKEY, "description": cls.DESCRIPTION}
                }
            },
        }


class InvalidPassword(LearningHouseSecurityException):
    INVALID_PASSWORD = "INVALID_PASSWORD"
    DESCRIPTION = "Invalid password"

    def __init__(self):
        super().__init__(self.DESCRIPTION, self.INVALID_PASSWORD)

    @classmethod
    def api_description(cls) -> Dict:
        return {
            "model": LearningHouseErrorMessage,
            "description": cls.DESCRIPTION,
            "content": {
                MIMETYPE_JSON: {
                    "example": {
                        "error": cls.INVALID_PASSWORD,
                        "description": cls.DESCRIPTION,
                    }
                }
            },
        }



from __future__ import annotations

from typing import Any, Dict, List

from pydantic import Field

from learninghouse.models.base import LHBaseModel


class LearningHouseVersions(LHBaseModel):
    service: str = Field(None, example="1.0.0")
    fastapi: str = Field(None, example="1.0.0")
    pydantic: str = Field(None, example="1.0.0")
    uvicorn: str = Field(None, example="1.0.0")
    sklearn: str = Field(None, example="1.0.0")
    numpy: str = Field(None, example="1.0.0")
    pandas: str = Field(None, example="1.0.0")
    jose: str = Field(None, example="1.0.0")
    passlib: str = Field(None, example="1.0.0")
    loguru: str = Field(None, example="1.0.0")

    @property
    def libraries_versions(self) -> str:
        return (
            f"Libraries FastAPI: {self.fastapi}, uvicorn: {self.uvicorn}, "
            + f"pydantic: {self.pydantic}, scikit-learn: {self.sklearn}, "
            + f"numpy: {self.numpy}, pandas: {self.pandas}, python-jose: {self.jose}, "
            + f"passlib: {self.passlib}, loguru: {self.loguru}"
        )


class LearningHouseErrorMessage(LHBaseModel):
    error: str
    description: str = ""


class LearningHouseValidationErrorMessage(LearningHouseErrorMessage):
    validations: List[Dict[str, Any]]

    @classmethod
    def from_error_message(
        cls, error_message: LearningHouseErrorMessage, validations: List[Dict[str, Any]]
    ) -> LearningHouseValidationErrorMessage:
        return cls(
            error=error_message.error,
            description=error_message.description,
            validations=validations,
        )



from fastapi import APIRouter

from learninghouse import versions
from learninghouse.api import auth, brain, sensor
from learninghouse.api.errors import LearningHouseSecurityException
from learninghouse.core.settings import service_settings
from learninghouse.models import LearningHouseVersions
from learninghouse.services.auth import authservice

api = APIRouter(
    prefix="/api",
    responses={
        LearningHouseSecurityException.STATUS_CODE: LearningHouseSecurityException.api_description()
    },
)

api.include_router(brain.router)
api.include_router(sensor.router)

api.include_router(auth.router)


@api.get("/mode", response_model=str, tags=["service"])
def get_mode():
    mode = service_settings().environment
    if authservice.is_initial_admin_password:
        mode = "initial"

    return mode


@api.get(
    "/versions",
    response_model=LearningHouseVersions,
    summary="Get versions",
    description="Get versions of the service and the used libraries",
    tags=["service"],
    responses={200: {"description": "Successfully retrieved versions"}},
)
def get_versions():
    return versions



from typing import List, Union

from fastapi import APIRouter, Depends, Path

from learninghouse.api.errors.auth import InvalidPassword
from learninghouse.models.auth import (
    APIKey,
    APIKeyInfo,
    APIKeyRequest,
    LoginRequest,
    PasswordRequest,
    Token,
    UserRole,
)
from learninghouse.services.auth import authservice


router = APIRouter(prefix="/auth", tags=["auth"])


@router.post(
    "/token",
    response_model=Token,
    responses={
        200: {"description": "Successfully retrieve token"},
        InvalidPassword.STATUS_CODE: InvalidPassword.api_description(),
    },
)
async def post_token(request: LoginRequest):
    return authservice.create_token(request.password)


@router.put("/token", response_model=Token)
async def put_token(refresh_token_jti: str = Depends(authservice.protect_refresh)):
    return authservice.refresh_token(refresh_token_jti)


@router.delete("/token", response_model=bool)
async def delete_token(
    refresh_token_jti: Union[str, None] = Depends(authservice.get_refresh)
):
    return authservice.revoke_refresh_token(refresh_token_jti)


router_protected = APIRouter(dependencies=[Depends(authservice.protect_admin)])


@router_protected.delete("/tokens", response_model=bool)
async def delete_tokens():
    return authservice.revoke_all_refresh_tokens()


@router_protected.put("/password", response_model=bool)
async def update_password(
    request: PasswordRequest, _=Depends(authservice.protect_admin)
):
    return authservice.update_password(request.old_password, request.new_password)


if not authservice.is_initial_admin_password:

    @router_protected.get("/apikeys", response_model=List[APIKeyInfo])
    async def list_api_keys():
        return authservice.list_api_keys()

    @router_protected.post("/apikey", response_model=APIKey)
    async def create_apikey(request: APIKeyRequest):
        return authservice.create_apikey(request)

    @router_protected.delete("/apikey/{description}", response_model=str)
    async def delete_apikey(
        description: str = Path(
            min_length=3,
            max_length=15,
            regex=r"^[A-Za-z]\w{1,13}[A-Za-z0-9]$",
            example="app_as_user",
        )
    ):
        return authservice.delete_apikey(description)


router.include_router(router_protected)


@router.get("/role", response_model=UserRole)
def role(user_role: UserRole = Depends(authservice.protect_user)):
    return user_role



from fastapi import APIRouter, Body, Depends, status

from learninghouse.api.errors.sensor import NoSensor, SensorExists
from learninghouse.models.sensor import Sensor, SensorDeleteResult, Sensors
from learninghouse.services.auth import authservice
from learninghouse.services.sensor import SensorConfigurationService

router = APIRouter(prefix="/sensor", tags=["sensor"])

router_usage = APIRouter(dependencies=[Depends(authservice.protect_user)])

router_admin = APIRouter(dependencies=[Depends(authservice.protect_admin)])


@router_usage.get(
    "s/configuration",
    response_model=Sensors,
    summary="Get all sensors configuration",
    description="Get all configured sensors.",
    responses={status.HTTP_200_OK: {"description": "All configured sensors"}},
)
async def get_sensors_configuration():
    return SensorConfigurationService.list_all()


@router_admin.get(
    "/{name}/configuration",
    response_model=Sensor,
    summary="Get sensor configuration",
    description="Get the current configuration of the given sensor",
    responses={
        status.HTTP_200_OK: {"description": "Configuration of the sensor"},
        NoSensor.STATUS_CODE: NoSensor.api_description(),
    },
)
async def get_sensor_configuration(name: str):
    return SensorConfigurationService.get(name)


@router_admin.post(
    "/configuration",
    response_model=Sensor,
    summary="Create a new sensor",
    description="Add a new sensor configuration.",
    status_code=status.HTTP_201_CREATED,
    responses={
        status.HTTP_201_CREATED: {"description": "Added new sensor"},
        SensorExists.STATUS_CODE: SensorExists.api_description(),
    },
)
async def post_sensor_configuration(sensor: Sensor):
    return SensorConfigurationService.create(sensor.name, sensor.typed)


@router_admin.put(
    "/{name}/configuration",
    response_model=Sensor,
    summary="Update a sensor",
    description="Update a existing sensor configuration.",
    responses={
        status.HTTP_200_OK: {"description": "Updated sensor"},
        NoSensor.STATUS_CODE: NoSensor.api_description(),
    },
)
async def put_sensor_configuration(name: str, sensor: Sensor = Body()):
    return SensorConfigurationService.update(name, sensor.typed)


@router_admin.delete(
    "/{name}/configuration",
    response_model=SensorDeleteResult,
    summary="Delete a sensor",
    description="Delete the configuration of a sensor.",
    responses={status.HTTP_200_OK: {"description": "DeleteSensor"}},
)
async def delete_sensor_configuration(name: str):
    return SensorConfigurationService.delete(name)


router.include_router(router_usage)
router.include_router(router_admin)



from fastapi import APIRouter, Depends, status

from learninghouse.api.errors.brain import (
    BrainNoConfiguration,
    BrainNotActual,
    BrainNotEnoughData,
    BrainNotTrained,
    BrainExists
)
from learninghouse.models.brain import (
    BrainConfiguration,
    BrainDeleteResult,
    BrainInfo,
    BrainInfos,
    BrainPredictionRequest,
    BrainPredictionResult,
    BrainTrainingRequest,
)
from learninghouse.services.auth import authservice
from learninghouse.services.brain import BrainService, BrainConfigurationService


router = APIRouter(prefix="/brain", tags=["brain"])

router_usage = APIRouter(dependencies=[Depends(authservice.protect_user)])

router_training = APIRouter(
    dependencies=[Depends(authservice.protect_trainer)])

router_admin = APIRouter(dependencies=[Depends(authservice.protect_admin)])


@router_usage.get(
    "s/info",
    response_model=BrainInfos,
    summary="Retrieve information",
    description="Retrieve all information about brains.",
    responses={
        200: {"description": "Information of all brains"},
    },
)
async def infos_get():
    return BrainService.list_all()


@router_usage.get(
    "/{name}/info",
    response_model=BrainInfo,
    summary="Retrieve information",
    description="Retrieve all information of a brain.",
    responses={
        200: {"description": "Information of the brain"},
        BrainNoConfiguration.STATUS_CODE: BrainNoConfiguration.api_description(),
    },
)
async def info_get(name: str):
    return BrainService.get_info(name)


@router_training.post(
    "/{name}/training",
    response_model=BrainInfo,
    summary="Train the brain again",
    description="After version updates train the brain with existing data.",
    responses={
        200: {"description": "Information of the trained brain"},
        BrainNotEnoughData.STATUS_CODE: BrainNotEnoughData.api_description(),
        BrainNoConfiguration.STATUS_CODE: BrainNoConfiguration.api_description(),
    },
)
async def training_post(name: str):
    return BrainService.request(name)


@router_training.put(
    "/{name}/training",
    response_model=BrainInfo,
    summary="Train the brain with new data",
    description="Train the brain with additional data.",
    responses={
        200: {"description": "Information of the trained brain"},
        BrainNotEnoughData.STATUS_CODE: BrainNotEnoughData.api_description(),
        BrainNoConfiguration.STATUS_CODE: BrainNoConfiguration.api_description(),
    },
)
async def training_put(name: str, request: BrainTrainingRequest):
    return BrainService.request(name, request.dependent_value, request.sensors_data)


@router_usage.post(
    "/{name}/prediction",
    response_model=BrainPredictionResult,
    summary="Prediction",
    description="Predict a new dataset with given brain.",
    responses={
        200: {"description": "Prediction result"},
        BrainNotActual.STATUS_CODE: BrainNotActual.api_description(),
        BrainNotTrained.STATUS_CODE: BrainNotTrained.api_description(),
    },
)
async def prediction_post(name: str, request_data: BrainPredictionRequest):
    return BrainService.prediction(name, request_data.dict())


@router_usage.get(
    "/{name}/configuration",
    response_model=BrainConfiguration,
    summary="Get configuration of a brain",
    description="Get the configuration of the specified brain",
    responses={
        status.HTTP_200_OK: {"description": "Configuration of the brain"},
        BrainNoConfiguration.STATUS_CODE: BrainNoConfiguration.api_description(),
    },
)
async def configuration_get(name: str):
    return BrainConfigurationService.get(name)


@router_admin.post(
    "/configuration",
    response_model=BrainConfiguration,
    summary="Create a new brain configuration",
    description="Put the configuration of a new brain",
    status_code=status.HTTP_201_CREATED,
    responses={
        status.HTTP_201_CREATED: {"description": "The new brain was created"},
        BrainExists.STATUS_CODE: BrainExists.api_description(),
    },
)
async def configuration_post(brain: BrainConfiguration):
    return BrainConfigurationService.create(brain)


@router_admin.put(
    "/{name}/configuration",
    response_model=BrainConfiguration,
    summary="Update brain configuration",
    description="Post the configuration to update the brain",
    responses={
        status.HTTP_200_OK: {"description": "The brain configuration was updated"},
        BrainNoConfiguration.STATUS_CODE: BrainNoConfiguration.api_description(),
    },
)
async def configuration_put(name: str, configuration: BrainConfiguration):
    return BrainConfigurationService.update(name, configuration)


@router_admin.delete(
    "/{name}/configuration",
    response_model=BrainDeleteResult,
    summary="Delete whole brain",
    responses={status.HTTP_200_OK: {
        "description": "Returns the name of the brain"}},
)
async def configuration_delete(name: str):
    return BrainConfigurationService.delete(name)


router.include_router(router_usage)
router.include_router(router_training)
router.include_router(router_admin)



from typing import Dict

from fastapi import status

from learninghouse import versions
from learninghouse.api.errors import LearningHouseException
from learninghouse.models import (
    LearningHouseErrorMessage,
    LearningHouseVersions,
)

MIMETYPE_JSON = "application/json"


class BrainNotTrained(LearningHouseException):
    STATUS_CODE = status.HTTP_404_NOT_FOUND
    NOT_TRAINED = "NOT_TRAINED"
    DESCRIPTION = "No trained brain with {name} found."

    def __init__(self, name: str):
        super().__init__(
            self.STATUS_CODE, self.NOT_TRAINED, self.DESCRIPTION.format(
                name=name)
        )

    @classmethod
    def api_description(cls) -> Dict:
        return {
            "model": LearningHouseErrorMessage,
            "description": "No trained brain with this name found.",
            "content": {
                MIMETYPE_JSON: {
                    "example": {
                        "error": cls.NOT_TRAINED,
                        "description": cls.DESCRIPTION,
                    }
                }
            },
        }


class BrainNotActual(LearningHouseException):
    STATUS_CODE = status.HTTP_428_PRECONDITION_REQUIRED
    NOT_ACTUAL = "NOT_ACTUAL"
    DESCRIPTION = (
        "The versions of trained brain {name} are "
        + "not the same as service. "
        + "Versions service: {versions}. "
        + "Versions brain: {brain_versions}. "
        + "Please train the brain again."
    )

    def __init__(self, name: str, brain_versions: LearningHouseVersions):
        super().__init__(
            self.STATUS_CODE,
            self.NOT_ACTUAL,
            self.DESCRIPTION.format(
                name=name, versions=versions, brain_versions=brain_versions
            ),
        )

    @classmethod
    def api_description(cls) -> Dict:
        return {
            "model": LearningHouseErrorMessage,
            "description": "If brain was not trained with actual versions "
            + "of service and libraries.",
            "content": {
                MIMETYPE_JSON: {
                    "example": {"error": cls.NOT_ACTUAL, "description": cls.DESCRIPTION}
                }
            },
        }


class BrainNotEnoughData(LearningHouseException):
    STATUS_CODE = status.HTTP_202_ACCEPTED
    NOT_ENOUGH_TRAINING_DATA = "NOT_ENOUGH_TRAINING_DATA"
    DESCRIPTION = (
        "Brain was not trained because at least 10 data points are needed. "
        + "But your new data point was saved."
    )

    def __init__(self):
        super().__init__(
            self.STATUS_CODE, self.NOT_ENOUGH_TRAINING_DATA, self.DESCRIPTION
        )

    @classmethod
    def api_description(cls) -> Dict:
        return {
            "model": LearningHouseErrorMessage,
            "description": "Response if there are not enough data points.",
            "content": {
                MIMETYPE_JSON: {
                    "example": {
                        "error": cls.NOT_ENOUGH_TRAINING_DATA,
                        "description": cls.DESCRIPTION,
                    }
                }
            },
        }


class BrainNoConfiguration(LearningHouseException):
    STATUS_CODE = status.HTTP_404_NOT_FOUND
    NO_CONFIGURATION = "NO_CONFIGURATION"
    DESCRIPTION = "No configuration for brain {name} found."

    def __init__(self, name: str):
        super().__init__(
            self.STATUS_CODE, self.NO_CONFIGURATION, self.DESCRIPTION.format(
                name=name)
        )

    @classmethod
    def api_description(cls) -> Dict:
        return {
            "model": LearningHouseErrorMessage,
            "description": "No brain configuration found.",
            "content": {
                MIMETYPE_JSON: {
                    "example": {
                        "error": cls.NO_CONFIGURATION,
                        "description": cls.DESCRIPTION,
                    }
                }
            },
        }


class BrainBadRequest(LearningHouseException):
    STATUS_CODE = status.HTTP_400_BAD_REQUEST
    BAD_REQUEST = "BAD_REQUEST"

    def __init__(self, description: str):
        super().__init__(self.STATUS_CODE, self.BAD_REQUEST, description)

    @classmethod
    def api_description(cls) -> Dict:
        return {
            "model": LearningHouseErrorMessage,
            "description": "Brain received a bad request.",
            "content": {
                MIMETYPE_JSON: {
                    "example": {"error": cls.BAD_REQUEST, "description": ""}
                }
            },
        }


class BrainExists(LearningHouseException):
    STATUS_CODE = status.HTTP_400_BAD_REQUEST
    EXISTS = "EXISTS"
    DESCRIPTION = "The existing brain {name} can not be recreated. Use PUT to update."

    def __init__(self, name: str):
        super().__init__(
            self.STATUS_CODE, self.EXISTS, self.DESCRIPTION.format(name=name)
        )

    @classmethod
    def api_description(cls) -> Dict:
        return {
            "model": LearningHouseErrorMessage,
            "description": "An existing brain can not be recreated. Use PUT to update.",
            "content": {
                MIMETYPE_JSON: {
                    "example": {"error": cls.EXISTS, "description": cls.DESCRIPTION}
                }
            },
        }



from fastapi import APIRouter
from fastapi.openapi.docs import get_swagger_ui_html
from starlette.responses import HTMLResponse

from learninghouse.core.settings import service_settings

settings = service_settings()

router = APIRouter(include_in_schema=False)


@router.get(settings.docs_url)
async def custom_swagger_ui_html() -> HTMLResponse:
    response = get_swagger_ui_html(
        openapi_url=settings.openapi_file,
        title=settings.title + " - Swagger UI",
        oauth2_redirect_url=None,
        swagger_js_url="/static/docs/swagger-ui-bundle.js",
        swagger_css_url="/static/docs/swagger-ui.css",
        swagger_favicon_url="/static/favicon.ico",
    )

    return response



from os import path

import setuptools

import versioneer

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt", "r", encoding="utf-8") as fh:
    requirements = fh.readlines()


packages = setuptools.find_packages(exclude=["tests"])
packages.append("learninghouse.static")
packages.append("learninghouse.static.docs")
if path.exists("learninghouse/ui"):
    packages.append("learninghouse.ui")
    packages.append("learninghouse.ui.assets")
    packages.append("learninghouse.ui.assets.fonts")
    packages.append("learninghouse.ui.assets.i18n")

setuptools.setup(
    name="learninghouse",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="learningHouse - Teach your smart home everything",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/LearningHouseService/learninghouse-monorepo",
    author="Johannes Ott",
    author_email="info@johannes-ott.net",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Topic :: Home Automation",
        "License :: OSI Approved :: MIT License",
        "Framework :: FastAPI",
        "Natural Language :: English",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.10",
    ],
    keywords="smart home, machine learning, house automation",
    packages=packages,
    include_package_data=True,
    python_requires=">=3.10, <4",
    install_requires=[req for req in requirements if req[:2] != "# "],
    entry_points={"console_scripts": ["learninghouse=learninghouse.service:run"]},
    project_urls={
        "Bug Reports": "https://github.com/LearningHouseService/learninghouse/issues"
    },
)



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Set, Tuple, Union

import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split

from learninghouse.models.sensor import Sensors

if TYPE_CHECKING:
    from learninghouse.models.brain import Brain


class DatasetPreprocessing:
    CATEGORICAL_KEY = "categorical"
    NUMERICAL_KEY = "numerical"

    @classmethod
    def sensorsconfig(cls) -> Tuple[List[str], List[str]]:
        sensors = Sensors.load_config()
        categoricals = sensors.categoricals
        numericals = sensors.numericals

        categoricals.append("month_of_year")
        numericals.append("day_of_month")
        categoricals.append("day_of_week")
        numericals.append("hour_of_day")
        numericals.append("minute_of_hour")

        return categoricals, numericals

    @staticmethod
    def add_time_information(data: Dict[str, Any]) -> Dict[str, Any]:
        if "timestamp" not in data:
            data["timestamp"] = datetime.now().timestamp()

        date = datetime.fromtimestamp(data["timestamp"])
        data["datetime"] = date.strftime("%Y-%m-%d %H:%M:%s")
        data["month_of_year"] = date.month
        data["day_of_month"] = date.day
        data["day_of_week"] = date.strftime("%A")
        data["hour_of_day"] = date.hour
        data["minute_of_hour"] = date.minute

        return data

    @classmethod
    def get_x_selected_and_numerical_columns(
        cls, brain: Brain, data: pd.DataFrame, only_features: bool
    ) -> Tuple[pd.DataFrame, List[str]]:
        categoricals, numericals = cls.sensorsconfig()

        categoricals = cls.columns_intersection(categoricals, data)

        used_columns = categoricals + cls.columns_intersection(numericals, data)

        if len(categoricals) > 0:
            x_temp = pd.get_dummies(data[used_columns], columns=categoricals)

            if only_features:
                features_in_dataframe = cls.columns_intersection(
                    x_temp, brain.dataset.features
                )

                x_selected = x_temp[features_in_dataframe]
            else:
                x_selected = x_temp
        else:
            if only_features:
                features_in_dataframe = cls.columns_intersection(
                    data, brain.dataset.features
                )

                x_selected = data[features_in_dataframe]
            else:
                x_selected = data[used_columns]

        x_selected = cls.sort_columns(x_selected)

        return x_selected, numericals

    @classmethod
    def prepare_training(
        cls, brain: Brain, data: pd.DataFrame, only_features: bool
    ) -> Tuple[Brain, pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        x_vector, numericals = cls.get_x_selected_and_numerical_columns(
            brain, data, only_features
        )

        y_vector = data[brain.configuration.name]

        if brain.configuration.dependent_encode:
            y_vector = brain.dataset.dependent_encoder.fit_transform(y_vector)

        x_train, x_test, y_train, y_test = train_test_split(
            x_vector, y_vector, test_size=brain.configuration.test_size, random_state=0
        )

        if only_features:
            numericals = cls.columns_intersection(numericals, brain.dataset.features)
        else:
            numericals = cls.columns_intersection(numericals, data)

        x_train = cls.transform_columns(
            brain.dataset.imputer.fit_transform, x_train, numericals
        )
        x_test = cls.transform_columns(
            brain.dataset.imputer.transform, x_test, numericals
        )

        x_train = cls.sort_columns(x_train)
        x_test = cls.sort_columns(x_test)

        return brain, x_train, x_test, y_train, y_test

    @classmethod
    def prepare_prediction(cls, brain: Brain, data: pd.DataFrame) -> pd.DataFrame:
        x_vector, numericals = cls.get_x_selected_and_numerical_columns(
            brain, data, True
        )

        numericals = cls.columns_intersection(brain.dataset.columns, numericals)

        missing_columns = set.difference(
            cls.set_of_columns(numericals), cls.set_of_columns(x_vector)
        )

        for missing_column in missing_columns:
            x_vector.insert(0, missing_column, [np.nan])

        x_vector = x_vector.reindex(columns=brain.dataset.columns, fill_value=0)
        x_vector = cls.sort_columns(x_vector)

        x_vector = cls.transform_columns(
            brain.dataset.imputer.transform, x_vector, numericals
        )

        return cls.sort_columns(x_vector)

    @staticmethod
    def transform_columns(
        func: Callable, data: pd.DataFrame, columns: List[str]
    ) -> pd.DataFrame():
        data_temp = data.copy()
        data_temp[columns] = func(data[columns])
        return data_temp

    @staticmethod
    def sort_columns(data: pd.DataFrame) -> pd.DataFrame:
        data_temp = data.copy()
        return data_temp.reindex(sorted(data.columns), axis=1)

    @classmethod
    def columns_intersection(
        cls,
        list_or_dataframe1: Union[pd.DataFrame, List[str]],
        list_or_dataframe2: Union[pd.DataFrame, List[str]],
    ) -> List[str]:
        set1 = cls.set_of_columns(list_or_dataframe1)
        set2 = cls.set_of_columns(list_or_dataframe2)
        return sorted(list(set.intersection(set1, set2)))

    @staticmethod
    def set_of_columns(list_or_dataframe: Union[pd.DataFrame, List[str]]) -> Set[str]:
        set_of_columns = None
        if isinstance(list_or_dataframe, pd.DataFrame):
            set_of_columns = set(list_or_dataframe.columns.values.tolist())
        elif isinstance(list_or_dataframe, list):
            set_of_columns = set(list_or_dataframe)

        return set_of_columns



from __future__ import annotations

import json
from datetime import datetime
from os import makedirs, path
from typing import Dict, List, Optional, Type

import joblib
from pydantic import Field, StrictBool, StrictFloat, StrictInt
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor

from learninghouse import versions
from learninghouse.api.errors import LearningHouseSecurityException
from learninghouse.api.errors.brain import BrainNotTrained
from learninghouse.core.settings import service_settings
from learninghouse.models import LearningHouseVersions
from learninghouse.models.base import DictModel, EnumModel, LHBaseModel
from learninghouse.models.preprocessing import DatasetConfiguration

settings = service_settings()


class BrainEstimatorType(EnumModel):
    """
    **LearningHouse Service** can predict values using an estimator. An estimator can be
    of type `classifier` which fits best for your needs if you have somekind of categorical
    output like in the darkness example true and false. If you want to predict a numerical
    value for example the setpoint of an heating equipment use the type `regressor` instead.
    """

    CLASSIFIER = "classifier", RandomForestClassifier
    REGRESSOR = "regressor", RandomForestRegressor

    def __init__(
        self,
        typed: str,
        estimator_class: Type[RandomForestClassifier] | Type[RandomForestRegressor]
    ):
        self._typed: str = typed
        self._estimator_class: Type[RandomForestClassifier] | Type[RandomForestRegressor] \
            = estimator_class

    @property
    def typed(self) -> str:
        return self._typed

    @property
    def estimator_class(
        self,
    ) -> Type[RandomForestClassifier] | Type[RandomForestRegressor]:
        return self._estimator_class


class BrainEstimatorConfiguration(LHBaseModel):
    """
    **LearningHouse Service** can predict values using an estimator.
    An estimator can be of type `classifier` which fits best for your
    needs if you have somekind of categorical output like in the
    darkness example true and false. If you want to predict a numerical
    value for example the setpoint of an heating equipment use the type
    `regressor` instead.

    For both types **learningHouse Service** uses a machine learning
    algorithm called random forest estimation. This algorithm builds
    a "forest" of decision trees with your `features` and takes the mean
    of the prediction of all of them to give you a best result. For more
    details see the API description of scikit-learn:

    | Estimator type | API Reference |
    |-----------------|-------------------|
    | RandomForestRegressor | \
        https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestRegressor.html#sklearn.ensemble.RandomForestRegressor |
    | RandomForestClassifier | \
        https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestClassifier.html#sklearn.ensemble.RandomForestClassifier |

    You can adjust the amount of decision trees by using `estimators`
    (default: 100) option. And the maximum depth of each tree by using
    `max_depth` (default: 5) option. Both options are optional. Try to
    resize this value to optimize the accuracy of your model.
    """  # pylint: disable=line-too-long

    typed: BrainEstimatorType = Field(
        None, example=BrainEstimatorType.CLASSIFIER)
    estimators: Optional[int] = Field(100, ge=100, le=1000)
    max_depth: Optional[int] = Field(5, ge=4, le=10)
    random_state: Optional[int] = Field(0)


class BrainConfiguration(LHBaseModel):
    """
    Estimator:
    See BrainEstimatorConfiguration

    Dependent variable:
    The `dependent` variable is the one that have to be in the training data and which is predicted by the trained brain.

    The `dependent` variable has to be a number. If it is not a number, but a string or boolean (true/false) like in the example. For this set `dependent_encode` to true.

    Test size:
    LearningHouse service only uses a part of your training data to train the brain. The other part specified by `test_size` will be used to score the accuracy of your brain.

    Give a percentage by using floating point numbers between 0.01 and 0.99 or a absolute number of data points by using integer numbers.

    For the beginning a `test_size` of 20 % (0.2) like the example should be fine.
    """  # pylint: disable=line-too-long

    name: str = Field(None, example="darkness")
    estimator: BrainEstimatorConfiguration
    dependent_encode: Optional[bool] = Field(False)
    test_size: Optional[float] = Field(0.2, gt=0.0, examples=[0.2, 20])

    @classmethod
    def from_json_file(cls, name: str) -> BrainConfiguration:
        filename = Brain.sanitize_filename(name, BrainFileType.CONFIG_FILE)

        with open(filename, "r", encoding="utf-8") as config_file:
            return BrainConfiguration(**json.load(config_file))

    @classmethod
    def json_config_file_exists(cls, name: str) -> bool:
        filename = Brain.sanitize_filename(name, BrainFileType.CONFIG_FILE)
        return path.exists(filename)

    def to_json_file(self, name: str) -> None:
        brainpath = Brain.sanitize_directory(name)
        makedirs(brainpath, exist_ok=True)

        filename = Brain.sanitize_filename(name, BrainFileType.CONFIG_FILE)

        self.write_to_file(filename, indent=4)


class BrainConfigurations(DictModel):
    """
    A dictionary of all available brain configurations.
    """

    root: Dict[str, BrainConfiguration]


class BrainInfo(LHBaseModel):
    """
    Information of the brain about configuration, training data size, score,
    trained at and versions.
    """

    name: str = Field(None, example="darkness")
    configuration: BrainConfiguration = Field(None)
    features: Optional[List[str]] = Field(
        None,
        example=["azimuth", "elevation", "rain_gauge",
                 "pressure_trend_1h_falling"],
    )
    training_data_size: int = Field(None, example=1234)
    score: float = Field(None, example=0.85)
    trained_at: Optional[datetime] = Field(
        None, example=datetime.now().strftime("%Y-%m-%dT%H:%M:%S.%f")
    )
    versions: LearningHouseVersions = Field(None, example=versions)
    actual_versions: bool = Field(True)


class BrainInfos(DictModel):
    """A dictionary of all available brains."""

    root: Dict[str, BrainInfo]


class BrainFileType(EnumModel):
    """
    Enumeration which holds the filetypes which are used for a brain
    """

    CONFIG_FILE = "config", "config.json"
    TRAINED_FILE = "trained", "trained.pkl"
    INFO_FILE = "info", "info.json"
    TRAINING_DATA_FILE = "data", "training_data.csv"
    ALL = "all", ""

    def __init__(self, typed: str, filename: str):
        # pylint: disable=super-init-not-called
        self._typed: str = typed
        self._filename: str = filename

    @property
    def typed(self) -> str:
        return self._typed

    @property
    def filename(self) -> str:
        return self._filename


class Brain:
    def __init__(self, name: str):
        self.name: str = name
        self.configuration: BrainConfiguration = BrainConfiguration.from_json_file(
            name)

        self.dataset: DatasetConfiguration = DatasetConfiguration(
            self.configuration)

        self._estimator: Optional[
            RandomForestClassifier | RandomForestRegressor
        ] = None

        self.score: Optional[float] = 0.0

        self.versions: LearningHouseVersions = versions

        self.trained_at: Optional[datetime] = None

    def estimator(self) -> RandomForestClassifier | RandomForestRegressor:
        if self._estimator is None:
            self._estimator = self.configuration.estimator.typed.estimator_class(
                n_estimators=self.configuration.estimator.estimators,
                max_depth=self.configuration.estimator.max_depth,
                random_state=self.configuration.estimator.random_state,
            )

        return self._estimator

    @property
    def actual_versions(self) -> bool:
        return self.versions == versions

    @property
    def info(self) -> BrainInfo:
        info = BrainInfo(
            name=self.name,
            configuration=self.configuration,
            features=self.dataset.features,
            training_data_size=self.dataset.data_size,
            score=self.score,
            trained_at=self.trained_at,
            versions=self.versions,
            actual_versions=self.actual_versions,
        )

        return info

    @classmethod
    def load_trained(cls, name: str) -> Brain:
        try:
            if not cls.is_trained(name):
                raise BrainNotTrained(name)

            filename = Brain.sanitize_filename(
                name, BrainFileType.TRAINED_FILE)

            loaded_brain = joblib.load(filename)

            return loaded_brain
        except (AttributeError, ModuleNotFoundError, KeyError) as exception:
            raise BrainNotTrained(name) from exception

    @classmethod
    def is_trained(cls, name: str) -> bool:
        filename = Brain.sanitize_filename(name, BrainFileType.TRAINED_FILE)
        return path.exists(filename)

    def store_trained(
        self, columns: List[str], training_data_size: int, score: float
    ) -> None:
        self.dataset.columns = columns
        self.dataset.data_size = training_data_size
        self.score = score
        self.trained_at = datetime.now()

        filename = Brain.sanitize_filename(
            self.name, BrainFileType.TRAINED_FILE)

        joblib.dump(self, filename)

        filename = Brain.sanitize_filename(self.name, BrainFileType.INFO_FILE)

        self.info.write_to_file(filename, 4)

    @staticmethod
    def sanitize_directory(name: str) -> str:
        brainpath = str(settings.brains_directory / name)

        fullpath = path.normpath(brainpath)

        if not fullpath.startswith(str(settings.brains_directory)):
            raise LearningHouseSecurityException(
                "Configuration file name breaks configuration directory"
            )

        return fullpath

    @classmethod
    def sanitize_filename(cls, name: str, filetype: BrainFileType) -> str:
        brainpath = cls.sanitize_directory(name)

        return path.normpath(path.join(brainpath, filetype.filename))


class BrainTrainingRequest(LHBaseModel):
    """
    For training with data send a PUT request to the service.

    You can send either a field `timestamp` with your dataset containing
    a UNIX-Timestamp or the service will add this information with its
    current time. The service generate some further time relevant fields
    inside the training dataset you can although use as `features`.
    These are `month_of_year`, `day_of_month`, `day_of_week`, `hour_of_day`
    and `minute_of_hour`

    If one of your sensors is not working at the moment and for this not
    sending a value the service will add a value by using the following
    rules. For `categorical data` all categorical columns will be set to
    zero. For `numerical data` the mean of all known training set values
    for this `feature` will be assumed.
    """

    dependent_value: StrictBool | StrictInt | StrictFloat = Field(
        None, example=True)
    sensors_data: Dict[str, StrictBool | StrictInt | StrictFloat | str | None] = \
        Field(None,
              example={
                  "azimuth": 321.4441223144531,
                  "elevation": -19.691608428955078,
                  "rain_gauge": 0.0,
                  "pressure": 971.0,
                  "pressure_trend_1h": "falling",
                  "temperature_outside": 23.0,
                  "temperature_trend_1h": "rising",
                  "light_state": False
              },
              )


class BrainPredictionRequest(DictModel):
    """
    For prediction send a POST request to the service.
    """

    root: Dict[str, StrictBool | StrictInt | StrictFloat | str | None] = Field(
        None,
        example={
            "azimuth": 321.4441223144531,
            "elevation": -19.691608428955078,
            "rain_gauge": 0.0,
            "pressure": 971.0,
            "pressure_trend_1h": "falling",
            "temperature_outside": 23.0,
            "temperature_trend_1h": "rising",
            "light_state": False,
        },
    )


class BrainPredictionResult(LHBaseModel):
    """
    The result of a prediction request."""

    brain: BrainInfo
    preprocessed: Dict[str, StrictBool | StrictInt | StrictFloat | str] = Field(
        None,
        example={
            "azimuth": 321.4441223144531,
            "elevation": -19.691608428955078,
            "rain_gauge": 0.0,
            "pressure_trend_1h_falling": 1,
        },
    )
    prediction: StrictBool | StrictInt | StrictFloat = Field(
        None, example=False)


class BrainDeleteResult(LHBaseModel):
    """
    The result of a delete request."""

    name: str = Field(None, example="darkness")



from datetime import datetime, timedelta
from functools import lru_cache
from typing import Dict, List, Tuple, Union

from fastapi import Security
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
from fastapi.security.api_key import APIKeyHeader, APIKeyQuery
from jose import JWTError, jwt

from learninghouse.api.errors import (
    LearningHouseSecurityException,
    LearningHouseUnauthorizedException,
)
from learninghouse.api.errors.auth import InvalidPassword
from learninghouse.core.logging import logger
from learninghouse.core.settings import service_settings
from learninghouse.models.auth import (
    APIKey,
    APIKeyInfo,
    APIKeyRequest,
    APIKeyRole,
    SecurityDatabase,
    Token,
    TokenPayload,
    UserRole,
)

settings = service_settings()

API_KEY_NAME = "X-LEARNINGHOUSE-API-KEY"

api_key_query = APIKeyQuery(name="api_key", auto_error=False)
api_key_header = APIKeyHeader(name=API_KEY_NAME, auto_error=False)
jwt_bearer = HTTPBearer(bearerFormat="JWT", auto_error=False)


INITIAL_PASSWORD_WARNING = """
In order to activate the service you have to replace the fallback password.

See https://github.com/LearningHouseService/learninghouse-monorepo/tree/main/learninghouse#fallback-password
"""


class AuthServiceInternal:
    def __init__(self):
        self.database = SecurityDatabase.load_or_write_default()
        self.refresh_tokens: Dict[str, datetime] = {}

    @property
    def is_initial_admin_password(self) -> bool:
        return self.database.initial_password

    def create_token(self, password: str) -> Token:
        if not self.database.authenticate_password(password):
            raise InvalidPassword()

        self.cleanup_refresh_tokens()
        token = self.create_new_token()

        logger.info("Admin user logged in sucessfully")

        return token

    def refresh_token(self, refresh_token_jti: str) -> Token:
        self.cleanup_refresh_tokens()

        if refresh_token_jti in self.refresh_tokens:
            del self.refresh_tokens[refresh_token_jti]

        token = self.create_new_token()

        logger.info("Admin token refreshed")

        return token

    def revoke_refresh_token(self, refresh_token_jti: Union[str, None]) -> bool:
        self.cleanup_refresh_tokens()

        if refresh_token_jti:
            if refresh_token_jti in self.refresh_tokens:
                del self.refresh_tokens[refresh_token_jti]

            logger.info("Logout admininstrator refresh token")

        return True

    def revoke_all_refresh_tokens(self) -> bool:
        self.refresh_tokens.clear()

        logger.warning("Revoked all refresh tokens")

        return True

    def cleanup_refresh_tokens(self):
        del_tokens = []
        for jti, expire in self.refresh_tokens.items():
            if expire < datetime.utcnow():
                del_tokens.append(jti)

        for jti in del_tokens:
            del self.refresh_tokens[jti]

    def create_new_token(self) -> Token:
        issuetime = datetime.utcnow()
        access_expire = issuetime + timedelta(minutes=1)
        access_payload = TokenPayload.create("admin", access_expire, issuetime)
        access_token = jwt.encode(access_payload.dict(), settings.jwt_secret, "HS256")

        refresh_expire = issuetime + timedelta(minutes=settings.jwt_expire_minutes)
        refresh_payload = TokenPayload.create("refresh", refresh_expire, issuetime)
        refresh_token = jwt.encode(refresh_payload.dict(), settings.jwt_secret, "HS256")

        self.refresh_tokens[refresh_payload.jti] = refresh_expire

        return Token(access_token=access_token, refresh_token=refresh_token)

    def update_password(self, old_password: str, new_password: str) -> bool:
        if not self.database.authenticate_password(old_password):
            raise InvalidPassword()

        self.database.update_password(new_password)
        self.database.write()
        self.refresh_tokens.clear()

        logger.info("New administration password set")

        return True

    def list_api_keys(self) -> List[APIKeyInfo]:
        return self.database.list_api_keys()

    def create_apikey(self, request: APIKeyRequest) -> APIKey:
        api_key = self.database.create_apikey(request)
        self.database.write()

        logger.info(f"New API key for {request.description} added")

        return api_key

    def delete_apikey(self, description: str) -> str:
        confirm = self.database.delete_apikey(description)
        self.database.write()

        logger.info(f"Removed API key for {description}.")

        return confirm

    async def protect_admin(
        self, credentials: HTTPAuthorizationCredentials = Security(jwt_bearer)
    ) -> UserRole:
        self.validate_credentials(credentials, True, "admin")
        return UserRole.ADMIN

    async def protect_refresh(
        self, credentials: HTTPAuthorizationCredentials = Security(jwt_bearer)
    ) -> str:
        _, jti = self.validate_credentials(credentials, True, "refresh")

        return jti

    async def get_refresh(
        self, credentials: HTTPAuthorizationCredentials = Security(jwt_bearer)
    ) -> Union[str, None]:
        is_valid, jti = self.validate_credentials(credentials, False, "refresh")

        return jti if is_valid else None

    async def protect_user(
        self,
        credentials: HTTPAuthorizationCredentials = Security(jwt_bearer),
        query: str = Security(api_key_query),
        header: str = Security(api_key_header),
    ) -> UserRole:
        role = self.is_admin_user_or_trainer(credentials, query, header)

        return role

    async def protect_trainer(
        self,
        credentials: HTTPAuthorizationCredentials = Security(jwt_bearer),
        query: str = Security(api_key_query),
        header: str = Security(api_key_header),
    ) -> UserRole:
        role = self.is_admin_user_or_trainer(credentials, query, header)

        if role.role not in ["admin", APIKeyRole.TRAINER.role]:
            raise LearningHouseUnauthorizedException()

        return role

    def is_admin_user_or_trainer(
        self, credentials: HTTPAuthorizationCredentials, query: str, header: str
    ) -> Union[UserRole, None]:
        role = None

        is_valid, _ = self.validate_credentials(credentials, False, "admin")

        if is_valid:
            role = UserRole.ADMIN
        else:
            key = query or header
            if not key:
                raise LearningHouseSecurityException("Invalid credentials")

            api_key_info = self.database.find_apikey_by_key(key)
            if not api_key_info:
                raise LearningHouseUnauthorizedException()

            role = UserRole.from_string(str(api_key_info.role))
        return role

    def validate_credentials(
        self,
        credentials: Union[HTTPAuthorizationCredentials, None],
        auto_error: bool,
        subject: str,
    ) -> Tuple[bool, Union[str, None]]:
        is_valid = True
        jti = None

        if credentials:
            if credentials.scheme != "Bearer":
                is_valid = False
                self.raise_error_conditionally(
                    "Invalid authentication scheme.", auto_error
                )

            verified, jti = self.verify_jwt(credentials.credentials, subject)

            if not verified:
                is_valid = False
                if auto_error:
                    raise LearningHouseUnauthorizedException()

        else:
            is_valid = False
            self.raise_error_conditionally("Invalid authorization code.", auto_error)

        return is_valid, jti

    @staticmethod
    def raise_error_conditionally(description: str, auto_error: bool):
        if auto_error:
            raise LearningHouseSecurityException(description)

    def verify_jwt(
        self, access_token: str, subject: str
    ) -> Tuple[bool, Union[str, None]]:
        verified = False
        jti = None
        try:
            payload = TokenPayload.parse_obj(
                jwt.decode(
                    access_token,
                    settings.jwt_secret,
                    "HS256",
                    subject=subject,
                    **settings.jwt_payload_claims,
                )
            )

            if subject == "refresh":
                verified = (
                    payload.jti in self.refresh_tokens
                    and self.refresh_tokens[payload.jti] > datetime.utcnow()
                )

                if not verified:
                    logger.error("No valid refresh token")
            else:
                verified = True

            jti = payload.jti
        except JWTError as exc:
            logger.info(exc)

        return verified, jti


@lru_cache()
def auth_service_cached() -> AuthServiceInternal:
    service = AuthServiceInternal()
    return service


authservice = auth_service_cached()



from learninghouse.api.errors.sensor import NoSensor, SensorExists
from learninghouse.models.sensor import (
    Sensor,
    SensorDeleteResult,
    Sensors,
    SensorType,
)


class SensorConfigurationService:
    @staticmethod
    def list_all() -> Sensors:
        return Sensors.load_config()

    @staticmethod
    def get(name: str) -> Sensor:
        sensors = Sensors.load_config()

        for sensor in sensors:  # type: Sensor
            if sensor.name == name:
                return sensor

        raise NoSensor(name)

    @staticmethod
    def create(name: str, typed: SensorType) -> Sensor:
        sensors = Sensors.load_config()

        for sensor in sensors:  # type: Sensor
            if sensor.name == name:
                raise SensorExists(name)

        new_sensor = Sensor(name=name, typed=typed)
        sensors.append(new_sensor)
        sensors.write_config()

        return new_sensor

    @staticmethod
    def update(name: str, typed: SensorType) -> Sensor:
        sensors = Sensors.load_config()
        for sensor in sensors:  # type: Sensor
            if sensor.name == name:
                sensor.typed = typed
                sensors.write_config()
                return sensor

        raise NoSensor(name)

    @staticmethod
    def delete(name: str) -> None:
        sensors = Sensors.load_config()

        for sensor in sensors:  # type: Sensor
            if sensor.name == name:
                sensors.remove(sensor)
                sensors.write_config()
                return SensorDeleteResult(name=name)

        raise NoSensor(name)



from functools import lru_cache
from pathlib import Path
from secrets import token_hex
from typing import Any, Dict, Optional, Union

from pydantic import DirectoryPath
from pydantic_settings import BaseSettings, SettingsConfigDict

from learninghouse import versions
from learninghouse.api.errors import (
    LearningHouseException,
    LearningHouseValidationError,
)
from learninghouse.core.logging import LoggingLevelEnum
from learninghouse.models.base import EnumModel

LICENSE_URL = "https://github.com/LearningHouseService/learninghouse/blob/main/LICENSE"


class ServiceSettings(BaseSettings):
    debug: Optional[bool] = False
    docs_url: str = "/docs"
    openapi_file: str = "/learninghouse_api.json"
    title: str = "learningHouse Service"

    host: str = "127.0.0.1"
    port: int = 5000

    workers: int = 5

    reload: bool = False
    base_url: str = ''

    environment: str = "production"

    config_directory: DirectoryPath = "./brains"

    logging_level: LoggingLevelEnum = LoggingLevelEnum.INFO

    jwt_secret: str = token_hex(16)
    jwt_expire_minutes: int = 10

    model_config = SettingsConfigDict(
        env_file=".env", env_prefix="learninghouse_", )

    @property
    def fastapi_kwargs(self) -> Dict[str, Any]:
        validation_error = LearningHouseValidationError
        return {
            "debug": self.debug,
            "title": self.title,
            "openapi_url": self.openapi_file,
            "docs_url": None,
            "redoc_url": None,
            "version": versions.service,
            "responses": {
                validation_error.STATUS_CODE: validation_error.api_description(),
                LearningHouseException.STATUS_CODE: LearningHouseException.api_description(),
            },
            "license_info": {"name": "MIT License", "url": LICENSE_URL},
        }

    @property
    def uvicorn_kwargs(self) -> Dict[str, Any]:
        kwargs = {
            "host": self.host,
            "port": self.port,
            "headers": [("server", f"LearningHouse Service {versions.service}")],
        }

        if self.reload:
            kwargs["reload"] = True
        else:
            kwargs["workers"] = self.workers

        return kwargs

    @property
    def brains_directory(self) -> Path:
        return Path(self.config_directory).absolute()

    @property
    def base_url_calculated(self) -> str:
        if self.base_url:
            base_url = self.base_url
        elif self.host in ("0.0.0.0", "127.0.0.1"):
            base_url = "http://localhost"
        else:
            base_url = f"http://{self.host}"

        return f"{base_url}:{self.port}"

    @property
    def documentation_url(self) -> Union[str, None]:
        documentation_url = None

        if self.docs_url is not None:
            documentation_url = self.base_url_calculated + self.docs_url

        return documentation_url

    @property
    def openapi_url(self) -> str:
        return self.base_url_calculated + self.openapi_file

    @property
    def jwt_payload_claims(self) -> Dict[str, str]:
        return {"audience": "LearningHouseAPI", "issuer": "LearningHouse Service"}


class DevelopmentSettings(ServiceSettings):
    environment: str = "development"
    debug: bool = True
    reload: bool = True
    title: str = "learningHouse Service - Development"
    api_key_required: bool = False

    model_config = SettingsConfigDict(
        env_file=".env", env_prefix="learninghouse_")


class ServiceEnvironment(EnumModel):
    PROD = "production", ServiceSettings
    DEV = "development", DevelopmentSettings

    def __init__(self, description: str, settings_class: ServiceSettings):
        # pylint: disable=super-init-not-called
        self._description: str = description
        self._settings_class: ServiceSettings = settings_class

    @property
    def description(self) -> str:
        return self._description

    @property
    def settings_class(self) -> ServiceSettings:
        return self._settings_class


class ServiceBaseSettings(BaseSettings):
    environment: Optional[ServiceEnvironment] = ServiceEnvironment.PROD

    model_config = SettingsConfigDict(
        env_file=".env", env_prefix="learninghouse_", extra="ignore")


@lru_cache()
def service_settings() -> ServiceSettings:
    environment = ServiceBaseSettings().environment
    return environment.settings_class()
